package interpreter

import interpreter.Result.*
import interpreter.StrictValue.*

class TestInterpreter extends munit.FunSuite:

  val strictEval = StrictInterpreter()
  val lazyEval = LazyInterpreter()

  test("simple1, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.simple1),
      Value(PairV(NumV(1), NumV(2)))
    )

  def assertLazyEval(expr: Expr, expected: Result[StrictValue]) =
    val strictCounter = StrictInterpreterWithCounter()
    val lazyCounter = LazyInterpreterWithCounter()
    strictCounter.topLevelEvaluate(expr)
    assertEquals(
      lazyCounter.topLevelEvaluate(expr),
      expected
    )
    assert(
      lazyCounter.counter > 0,
      f"Lazy interpreter does not call evaluate()"
    )
    assert(
      lazyCounter.counter < strictCounter.counter,
      f"Lazy interpreter took ${lazyCounter.counter} steps, which is not fewer than the strict one (${strictCounter.counter})."
    )

  test("simple1, lazy (1pt)"):
    assertEquals(
      LazyInterpreter().topLevelEvaluate(ExprHelpers.simple1),
      Value(PairV(NumV(1), NumV(2)))
    )

  test("simple2, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.simple2),
      Value(NumV(3))
    )

  test("simple2, lazy (1pt)"):
    assertLazyEval(ExprHelpers.simple2, Value(NumV(3)))

  test("false conditional, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.falseConditional),
      Value(PairV(
        NumV(0),
        NumV(1)
      ))
    )

  test("false conditional, lazy (1pt)"):
    assertLazyEval(
      ExprHelpers.falseConditional,
      Value(PairV(
        NumV(0),
        NumV(1)
      ))
    )

  test("true conditional, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.trueConditional),
      Value(PairV(
        NumV(0),
        NumV(0)
      ))
    )

  test("true conditional, lazy (1pt)"):
    assertEquals(
      lazyEval.topLevelEvaluate(ExprHelpers.trueConditional),
      Value(PairV(
        NumV(0),
        NumV(0)
      ))
    )

  test("forced division by zero error, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.divByZero),
      InterpreterError(
        ErrorType.DivisionByZero
      )
    )

  test("forced division by zero error, lazy (1pt)"):
    assertEquals(
      lazyEval.topLevelEvaluate(ExprHelpers.divByZero),
      InterpreterError(
        ErrorType.DivisionByZero
      )
    )

  test("unforced division by zero still produces an error, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.divByZero2),
      InterpreterError(
        ErrorType.DivisionByZero
      )
    )

  test("unforced division by zero produces no error, lazy (1pt)"):
    assertLazyEval(ExprHelpers.divByZero2, Value(NumV(1)))

  test("unbound variable error, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.unboundVar),
      InterpreterError(
        ErrorType.UnboundVariable("z")
      )
    )

  test("unbound variable error, lazy (1pt)"):
    assertEquals(
      lazyEval.topLevelEvaluate(ExprHelpers.unboundVar),
      InterpreterError(
        ErrorType.UnboundVariable("z")
      )
    )

  test("NumOpOnNonNum error, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.numOpOnNonNum),
      InterpreterError(
        ErrorType.NumOpOnNonNum
      )
    )

  test("NumOpOnNonNum error, lazy (1pt)"):
    assertLazyEval(
      ExprHelpers.numOpOnNonNum,
      InterpreterError(
        ErrorType.NumOpOnNonNum
      )
    )

  test("PairOpOnNonPair error, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.pairOpOnNonPair),
      InterpreterError(
        ErrorType.PairOpOnNonPair
      )
    )

  test("PairOpOnNonPair error, lazy (1pt)"):
    assertEquals(
      lazyEval.topLevelEvaluate(ExprHelpers.pairOpOnNonPair),
      InterpreterError(
        ErrorType.PairOpOnNonPair
      )
    )

  test("IfCondNotNum error, strict (1pt)"):
    assertEquals(
      strictEval.topLevelEvaluate(ExprHelpers.ifCondNotNum),
      InterpreterError(
        ErrorType.IfCondNotNum
      )
    )

  test("IfCondNotNum error, lazy (1pt)"):
    assertLazyEval(
      ExprHelpers.ifCondNotNum,
      InterpreterError(
        ErrorType.IfCondNotNum
      )
    )
